import pytest


def test_sum():
    assert 1 + 1 == 2


def test_fail():
    assert 1 + 1 == 3


@pytest.mark.skip
def test_mark_skip():
    assert 1 + 1 == 3


@pytest.mark.parametrize(
    "a, b, result",
    [
        (1, 1, 2),
        (1, 2, 3),
    ],
)
def test_mark(a, b, result):
    assert a + b == result



from textual.containers import Horizontal
from textual.widgets import Button

from rich.text import Text
from textual_slidecontainer import SlideContainer


class TreeFilter(SlideContainer):
    # file_path_to_preview: reactive[Path | None] = reactive(None, init=False)
    # test_start_line_no: reactive[int] = reactive(-1, init=False)

    def __init__(self, *args, **kwargs):
        super().__init__(
            slide_direction="down",
            floating=False,
            start_open=False,
            duration=0.5,
            *args,
            **kwargs,
        )

    def compose(self):
        with Horizontal():
            yield Button(
                label=Text.from_markup("Favourites: :star:"),
                id="button_filter_favourites",
                variant="success",
            )
            yield Button(
                label=Text.from_markup("Passed: :white_check_mark:"),
                id="button_filter_passed",
                variant="success",
            )
            yield Button(
                label=Text.from_markup("Failed: :x:"),
                id="button_filter_failed",
                variant="success",
            )
            yield Button(
                label=Text.from_markup("Skipped: [on yellow]:next_track_button: [/]"),
                id="button_filter_skipped",
                variant="success",
            )



from pathlib import Path

from textual.reactive import reactive
from textual.widgets import TextArea
from textual_slidecontainer import SlideContainer

from ayu.utils import get_preview_test


class CodePreview(SlideContainer):
    file_path_to_preview: reactive[Path | None] = reactive(None, init=False)
    test_start_line_no: reactive[int] = reactive(-1, init=False)

    def __init__(self, *args, **kwargs):
        super().__init__(
            slide_direction="up",
            floating=False,
            start_open=False,
            duration=0.5,
            *args,
            **kwargs,
        )

    def compose(self):
        yield TextArea.code_editor(
            "Please select a test",
            id="textarea_preview",
            language="python",
            read_only=True,
        )

    def watch_file_path_to_preview(self):
        self.border_title = self.file_path_to_preview.as_posix()

    def watch_test_start_line_no(self):
        if self.test_start_line_no == -1:
            self.app.query_one("#textarea_preview").text = "Please select a test"
        else:
            content = get_preview_test(
                file_path=self.file_path_to_preview,
                start_line_no=self.test_start_line_no,
            )
            self.app.query_one(
                "#textarea_preview", TextArea
            ).line_number_start = self.test_start_line_no
            self.app.query_one("#textarea_preview", TextArea).text = content



from enum import Enum
from pathlib import Path
import subprocess


class NodeType(str, Enum):
    DIR = "DIR"
    MODULE = "MODULE"
    CLASS = "CLASS"
    FUNCTION = "FUNCTION"
    COROUTINE = "COROUTINE"


class EventType(str, Enum):
    COLLECTION = "COLLECTION"
    SCHEDULED = "SCHEDULED"
    OUTCOME = "OUTCOME"
    REPORT = "REPORT"


class TestOutcome(str, Enum):
    PASSED = "PASSED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    QUEUED = "QUEUED"


def run_test_collection():
    if Path.cwd().name == "ayu":
        command = "pytest --co".split()
    else:
        command = "uv run --with ../ayu pytest --co".split()
    subprocess.run(
        command,
        # ["pytest", "--co"],
        # ["uv", "run", "--with", "../ayu", "-U", "pytest", "--co"],
        capture_output=True,
    )


def run_all_tests(tests_to_run: list[str] | None = None):
    if Path.cwd().name == "ayu":
        command = "python -m pytest".split()
    else:
        command = "uv run --with ../ayu pytest".split()
        # command = "python -m pytest".split()

    if tests_to_run:
        command.extend(tests_to_run)

    subprocess.run(
        command,
        # ["pytest", "--co"],
        # ["uv", "run", "--with", "../ayu", "-U", "pytest", "--co"],
        capture_output=True,
    )


def get_nice_tooltip(node_data: dict) -> str | None:
    tooltip_str = ""
    # tooltip_str = f"{node_data['name'].replace("[", "\["):^20}\n"
    # tooltip_str += f"[red strike]{node_data['name'].replace('[', '\['):^20}[/]\n"
    #
    # status = node_data["status"].replace("[", "\[")
    # tooltip_str += f"\n[yellow]{status}[/]\n\n"
    return tooltip_str


def get_preview_test(file_path: str, start_line_no: int) -> str:
    with open(Path(file_path), "r") as file:
        file_lines = file.readlines()
        last_line_is_blank = False
        end_line_no = None
        for line_no, line in enumerate(file_lines[start_line_no:], start=start_line_no):
            if not line.strip():
                last_line_is_blank = True
                continue
            if (
                line.strip().startswith(("def ", "class ", "async def ", "@"))
                and last_line_is_blank
            ):
                end_line_no = line_no - 1
                break
            last_line_is_blank = False
        return "".join(file_lines[start_line_no:end_line_no]).rstrip()



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ayu.app import AyuApp
from textual import work
from textual.reactive import reactive
from textual.binding import Binding
from textual.widgets import Tree
from textual.widgets.tree import TreeNode
from rich.text import Text

from ayu.utils import (
    EventType,
    NodeType,
    TestOutcome,
    get_nice_tooltip,
    run_test_collection,
)
from ayu.constants import OUTCOME_SYMBOLS


class TestTree(Tree):
    app: "AyuApp"
    BINDINGS = [
        Binding("r", "collect_tests", "Refresh"),
        Binding("j,down", "cursor_down"),
        Binding("k,up", "cursor_up"),
        Binding("f", "mark_test_as_fav", "⭐ Mark"),
    ]
    show_root = False
    auto_expand = True
    guide_depth = 2

    counter_queued: reactive[int] = reactive(0)
    counter_passed: reactive[int] = reactive(0)
    counter_failed: reactive[int] = reactive(0)
    counter_skipped: reactive[int] = reactive(0)
    counter_marked: reactive[int] = reactive(0)

    filtered_data_test_tree: reactive[dict] = reactive({}, init=False)
    filtered_counter_total_tests: reactive[int] = reactive(0, init=False)
    filter: reactive[dict] = reactive(
        {
            "show_favourite": True,
            "show_failed": True,
            "show_skipped": True,
            "show_passed": True,
        },
        init=False,
    )

    def on_mount(self):
        self.app.dispatcher.register_handler(
            event_type=EventType.SCHEDULED,
            handler=lambda data: self.mark_tests_as_running(data),
        )
        self.app.dispatcher.register_handler(
            event_type=EventType.OUTCOME,
            handler=lambda data: self.update_test_outcome(data),
        )

        self.action_collect_tests()

        return super().on_mount()

    def watch_filtered_counter_total_tests(self):
        self.update_border_title()

    def watch_filtered_data_test_tree(self):
        if self.filtered_data_test_tree:
            self.build_tree()

    @work(thread=True)
    def action_collect_tests(self):
        run_test_collection()

    def build_tree(self):
        self.clear()
        self.reset_status_counters()
        self.counter_marked = 0
        self.update_tree(tree_data=self.filtered_data_test_tree)

    def update_tree(self, *, tree_data: dict[Any, Any]):
        parent = self.root

        def add_children(child_list: list[dict[Any, Any]], parent_node: TreeNode):
            for child in child_list:
                if child["children"]:
                    new_node = parent_node.add(
                        label=child["name"], data=child, expand=True
                    )
                    # Update labels?
                    add_children(child_list=child["children"], parent_node=new_node)
                else:
                    new_node = parent_node.add_leaf(label=child["name"], data=child)

        for key, value in tree_data.items():
            if isinstance(value, dict) and "children" in value and value["children"]:
                node: TreeNode = parent.add(key, data=value)
                self.select_node(node)
                add_children(value["children"], node)
            else:
                parent.add_leaf(key, data=key)

    def update_test_outcome(self, test_result: dict):
        for node in self._tree_nodes.values():
            if node.data and (node.data["nodeid"] == test_result["nodeid"]):
                outcome = test_result["outcome"]
                # node.label = f"{node.label} {OUTCOME_SYMBOLS[outcome]}"
                node.data["status"] = outcome
                node.label = self.update_test_node_label(node=node)
                self.counter_queued -= 1
                match outcome:
                    case TestOutcome.PASSED:
                        self.counter_passed += 1
                    case TestOutcome.FAILED:
                        self.counter_failed += 1
                    case TestOutcome.SKIPPED:
                        self.counter_skipped += 1

                self.update_collapse_state_on_test_run(node=node)

    def update_collapse_state_on_test_run(self, node: TreeNode):
        if node.parent.data["type"] == NodeType.CLASS:
            self.update_collapse_state_on_test_run(node=node.parent)
        if self.all_child_tests_passed(parent=node.parent):
            # node.parent.collapse()
            node.parent.label = self.update_mod_class_node_label(node=node.parent)
            node.parent.label = f"[green]{node.parent.label}[/]"
        else:
            node.parent.expand_all()
            node.parent.label = self.update_mod_class_node_label(node=node.parent)
            node.parent.label = f"[red]{node.parent.label}[/]"

    def all_child_tests_passed(self, parent: TreeNode):
        return all(
            [
                self.all_child_tests_passed(parent=child)
                if child.data["type"] == NodeType.CLASS
                else child.data["status"] in [TestOutcome.PASSED, TestOutcome.QUEUED]
                for child in parent.children
            ]
        )

    def reset_status_counters(self) -> None:
        self.counter_queued = 0
        self.counter_passed = 0
        self.counter_skipped = 0
        self.counter_failed = 0

    def mark_tests_as_running(self, nodeids: list[str]) -> None:
        self.root.expand_all()
        self.reset_status_counters()
        for node in self._tree_nodes.values():
            if node.data and (node.data["nodeid"] in nodeids):
                node.data["status"] = TestOutcome.QUEUED
                node.label = self.update_test_node_label(node=node)
                self.counter_queued += 1

    def on_tree_node_selected(self, event: Tree.NodeSelected):
        ...
        # Run Test

    def action_mark_test_as_fav(
        self, node: TreeNode | None = None, parent_val: bool | None = None
    ):
        # mark filtered tree as fav
        # self.app.data_test_tree['ayu']["children"][0]["favourite"] = True

        # if no node given, select node under cursor
        if node is None:
            node = self.cursor_node

        if parent_val is None:
            parent_val = not node.data["favourite"]

        # mark all childs the same as parent
        if node.children:
            node.data["favourite"] = parent_val
            node.label = self.update_test_node_label(node=node)
            for child in node.children:
                self.action_mark_test_as_fav(node=child, parent_val=parent_val)
        else:
            if node.data["favourite"] != parent_val:
                self.counter_marked += 1 if parent_val else -1
            node.data["favourite"] = parent_val
            node.label = self.update_test_node_label(node=node)

        if not node.data["favourite"]:
            parent_node = node.parent
            while parent_node.data is not None:
                parent_node.data["favourite"] = node.data["favourite"]
                parent_node.label = self.update_test_node_label(node=parent_node)
                parent_node = parent_node.parent

    def update_mod_class_node_label(self, node: TreeNode) -> str:
        counter_childs_tests = len(
            [
                child
                for child in node.children
                if child.data["type"] in [NodeType.FUNCTION, NodeType.COROUTINE]
            ]
        )
        # Misses Class Case
        counter_childs_test_passed = len(
            [
                child
                for child in node.children
                if child.data["status"] == TestOutcome.PASSED
            ]
        )
        fav_substring = "⭐ " if node.data["favourite"] else ""
        if counter_childs_test_passed == counter_childs_tests:
            node.collapse()

        return f"{fav_substring}{node.data['name']} ({counter_childs_test_passed}/{counter_childs_tests})"

    def update_test_node_label(self, node: TreeNode) -> str:
        fav_substring = "⭐ " if node.data["favourite"] else ""
        status_substring = (
            f" {OUTCOME_SYMBOLS[node.data['status']]}" if node.data["status"] else ""
        )

        return f"{fav_substring}{node.data['name']}{status_substring}"

    def on_mouse_move(self):
        return
        if self.hover_line != -1:
            data = self._tree_lines[self.hover_line].node.data
            self.tooltip = get_nice_tooltip(node_data=data)

    def watch_counter_queued(self):
        self.update_border_title()

    def watch_counter_passed(self):
        self.update_border_title()

    def watch_counter_failed(self):
        self.update_border_title()

    def watch_counter_skipped(self):
        self.update_border_title()

    def watch_counter_marked(self):
        self.update_border_title()

    def update_border_title(self):
        symbol = "hourglass_not_done" if self.counter_queued > 0 else "hourglass_done"
        tests_to_run = (
            self.app.counter_total_tests
            if not self.counter_marked
            else f":star: {self.counter_marked}/{self.app.counter_total_tests}"
        )

        self.border_title = Text.from_markup(
            f" :{symbol}: {self.counter_queued} | :x: {self.counter_failed}"
            + f" | :white_check_mark: {self.counter_passed} | :next_track_button: {self.counter_skipped}"
            + f" | Tests to run {tests_to_run} "
        )

    @property
    def marked_tests(self):
        marked_tests = []
        for node in self._tree_nodes.values():
            if (
                node.data
                and (node.data["type"] in [NodeType.FUNCTION, NodeType.COROUTINE])
                and node.data["favourite"]
            ):
                marked_tests.append(node.data["nodeid"])
        return marked_tests

    def reset_test_results(self):
        self.reset_status_counters()
        for node in self._tree_nodes.values():
            if (
                node.data
                and (node.data["type"] in [NodeType.FUNCTION, NodeType.COROUTINE])
                and node.data["status"]
            ):
                node.data["status"] = ""
                node.label = self.update_test_node_label(node=node)
            elif node.data and (node.data["type"] in [NodeType.MODULE, NodeType.CLASS]):
                node.label = self.update_test_node_label(node=node)



from typing import Callable
from collections import defaultdict

from websockets.asyncio.server import serve
from websockets.asyncio.client import connect
from websockets.exceptions import ConnectionClosedOK
import asyncio

from ayu.constants import WEB_SOCKET_HOST, WEB_SOCKET_PORT
from ayu.classes.event import Event
from ayu.utils import EventType


class EventDispatcher:
    def __init__(self, host: str = WEB_SOCKET_HOST, port: int = WEB_SOCKET_PORT):
        self.host = host
        self.port = port
        self.running = False
        self.server = None
        self.event_handler: defaultdict[EventType | None, list] = defaultdict(list)

        self.data = ""

    # Handler
    async def handler(self, websocket):
        while True:
            try:
                msg = await websocket.recv()
            except ConnectionClosedOK:
                break

            if self.event_handler:
                event = Event.deserialize(msg)
                event_type = event.event_type
                event_payload = event.event_payload
                match event_type:
                    case EventType.COLLECTION:
                        handlers = self.event_handler[EventType.COLLECTION]
                    case EventType.OUTCOME:
                        handlers = self.event_handler[EventType.OUTCOME]
                    case EventType.REPORT:
                        handlers = self.event_handler[EventType.REPORT]
                    case EventType.SCHEDULED:
                        handlers = self.event_handler[EventType.SCHEDULED]

            if handlers:
                for handler in handlers:
                    handler(event_payload)

            self.data = msg

    def register_handler(self, event_type: EventType, handler: Callable):
        self.event_handler[event_type].append(handler)

    def unregister_handler(self, event_type: EventType):
        # with asyncio.Lock():
        self.event_handler.pop(event_type)

    # Start Websocket Server
    async def start(self):
        await self.start_socket_server()

    def stop(self):
        asyncio.get_running_loop().create_future()

    async def start_socket_server(self):
        self.server = await serve(self.handler, self.host, self.port)
        print("started")
        await self.server.wait_closed()

    def get_data(self):
        return self.data


# send events
async def send_event(
    event: Event, host: str = WEB_SOCKET_HOST, port: int = WEB_SOCKET_PORT
):
    uri = f"ws://{host}:{port}"

    async with connect(uri) as websocket:
        await websocket.send(message=event.serialize())


async def check_connection(host: str = WEB_SOCKET_HOST, port: int = WEB_SOCKET_PORT):
    uri = f"ws://{host}:{port}"
    async with connect(uri) as websocket:
        await websocket.close()



WEB_SOCKET_HOST = "localhost"
WEB_SOCKET_PORT = 1337

OUTCOME_SYMBOLS = {
    "PASSED": ":white_check_mark:",
    "FAILED": ":cross_mark:",
    "SKIPPED": "[on yellow]:next_track_button: [/]",
    "QUEUED": ":hourglass_not_done:",
}



from pathlib import Path
from textual import work, on
from textual.app import App
from textual.reactive import reactive
from textual.events import Key
from textual.widgets import Log, Header, Footer, Collapsible, Tree
from textual.containers import Horizontal, Vertical
# from textual_slidecontainer import SlideContainer

from ayu.event_dispatcher import EventDispatcher
from ayu.utils import EventType, NodeType, run_all_tests
from ayu.widgets.navigation import TestTree
from ayu.widgets.code_preview import CodePreview
from ayu.widgets.filter import TreeFilter


class AyuApp(App):
    CSS_PATH = Path("assets/ayu.tcss")
    TOOLTIP_DELAY = 0.5

    data_test_tree: reactive[dict] = reactive({}, init=False)
    counter_total_tests: reactive[int] = reactive(0, init=False)

    def __init__(self, *args, **kwargs):
        self.dispatcher = None
        super().__init__(*args, **kwargs)

    def compose(self):
        yield Header()
        yield Footer()
        outcome_log = Log(highlight=True, id="log_outcome")
        outcome_log.border_title = "Outcome"
        report_log = Log(highlight=True, id="log_report")
        report_log.border_title = "Report"
        collection_log = Log(highlight=True, id="log_collection")
        collection_log.border_title = "Collection"
        with Horizontal():
            with Vertical():
                yield TestTree(label="Tests").data_bind(
                    filtered_data_test_tree=AyuApp.data_test_tree,
                    filtered_counter_total_tests=AyuApp.counter_total_tests,
                )
                yield TreeFilter()
            with Vertical():
                yield CodePreview()
                with Collapsible(title="Outcome", collapsed=True):
                    yield outcome_log
                with Collapsible(title="Report", collapsed=True):
                    yield report_log
                with Collapsible(title="Collection", collapsed=True):
                    yield collection_log

    async def on_load(self):
        self.start_socket()

    def on_mount(self):
        self.dispatcher.register_handler(
            event_type=EventType.OUTCOME,
            handler=lambda msg: self.update_outcome_log(msg),
        )
        self.dispatcher.register_handler(
            event_type=EventType.REPORT, handler=lambda msg: self.update_report_log(msg)
        )
        self.app.dispatcher.register_handler(
            event_type=EventType.COLLECTION,
            handler=lambda data: self.update_app_data(data),
        )

    def update_app_data(self, data):
        self.data_test_tree = data["tree"]
        self.counter_total_tests = data["meta"]["test_count"]

    @work(exclusive=True)
    async def start_socket(self):
        self.dispatcher = EventDispatcher()
        self.notify("Websocket Started", timeout=1)
        await self.dispatcher.start()

    def on_key(self, event: Key):
        if event.key == "ctrl+j":
            self.run_test()
        if event.key == "w":
            self.notify(f"{self.workers}")
        if event.key == "s":
            self.query_one(CodePreview).toggle()
            self.query_one(TreeFilter).toggle()
        if event.key == "c":
            self.query_one(TestTree).reset_test_results()
            for log in self.query(Log):
                log.clear()

    @on(Tree.NodeHighlighted)
    def update_test_preview(self, event: Tree.NodeHighlighted):
        self.query_one(CodePreview).file_path_to_preview = Path(event.node.data["path"])
        if event.node.data["type"] in [
            NodeType.FUNCTION,
            NodeType.COROUTINE,
            NodeType.CLASS,
        ]:
            self.query_one(CodePreview).test_start_line_no = event.node.data["lineno"]
        else:
            self.query_one(CodePreview).test_start_line_no = -1

    @work(thread=True)
    def run_test(self):
        run_all_tests(tests_to_run=self.query_one(TestTree).marked_tests)

    def update_outcome_log(self, msg):
        self.query_one("#log_outcome", Log).write_line(f"{msg}")

    def update_report_log(self, msg):
        self.query_one("#log_report", Log).write_line(f"{msg}")

    def watch_data_test_tree(self):
        self.query_one("#log_collection", Log).write_line(f"{self.data_test_tree}")


# https://watchfiles.helpmanual.io



def test_sum():
    assert 1 + 1 == 2


def test_fail():
    assert False


class TestClass:
    def test_class_sum(self):
        assert 1 + 1 == 2

    def test_class_fail(self):
        assert False



from __future__ import annotations
from dataclasses import dataclass
import json

from ayu.utils import EventType


@dataclass
class Event:
    event_type: EventType
    event_payload: dict | list

    def serialize(self) -> str:
        return json.dumps(
            {"type": self.event_type, "payload": self.event_payload}, indent=4
        )

    @classmethod
    def deserialize(cls, json_str: str) -> Event:
        event_dict = json.loads(json_str)
        return cls(event_type=event_dict["type"], event_payload=event_dict["payload"])



import os
import asyncio
from typing import Any
import pytest
from pytest import Config, TestReport, Session, Item, Class, Function
from _pytest.terminal import TerminalReporter
from _pytest.nodes import Node

from ayu.event_dispatcher import send_event, check_connection
from ayu.classes.event import Event
from ayu.utils import EventType, TestOutcome


def pytest_addoption(parser) -> None:
    parser.addoption(
        "--disable-ayu",
        "--da",
        action="store_true",
        default=False,
        help="Enable Ayu plugin functionality",
    )


@pytest.hookimpl(trylast=True)
def pytest_configure(config: Config) -> None:
    if not config.getoption("--disable-ayu"):
        config.pluginmanager.register(Ayu(config), "ayu_plugin")


class Ayu:
    def __init__(self, config: Config):
        self.config = config
        self.connected = False
        try:
            asyncio.run(check_connection())
            print("connected")
            self.connected = True
        except OSError:
            self.connected = False
            print("Websocket not connected")

    # build test tree
    def pytest_collection_finish(self, session: Session):
        print("Connected to Ayu")
        if self.connected:
            if session.config.getoption("--collect-only"):
                tree = build_dict_tree(items=session.items)
                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.COLLECTION,
                            event_payload=tree,
                        )
                    )
                )
            else:
                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.SCHEDULED,
                            event_payload=[item.nodeid for item in session.items],
                        )
                    )
                )
        return

    # gather status updates during run
    # @pytest.hookimpl(trylast=True)
    def pytest_runtest_logreport(self, report: TestReport):
        if self.config.pluginmanager.hasplugin("xdist") and (
            "PYTEST_XDIST_WORKER" not in os.environ
        ):
            return

        is_relevant = (report.when == "call") or (
            (report.when == "setup")
            and (report.outcome.upper() in [TestOutcome.FAILED, TestOutcome.SKIPPED])
        )

        if self.connected and is_relevant:
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.OUTCOME,
                        event_payload={
                            "nodeid": report.nodeid,
                            "outcome": report.outcome.upper(),
                            "report": f"{report}",
                        },
                    )
                )
            )

    # summary after run for each tests
    @pytest.hookimpl(tryfirst=True)
    def pytest_terminal_summary(self, terminalreporter: TerminalReporter):
        if self.connected:
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.REPORT,
                        event_payload={
                            "report": f"{terminalreporter.stats.get('', [])}",
                        },
                    )
                )
            )
        return
        for report in terminalreporter.stats.get("", []):
            if report.when == "teardown":
                print("## Summary ##")
                print(f"when: {report.when}")
                print(f"cap: {report.caplog}")
                print(f"long: {report.longreprtext}")
                print(f"dur: {report.duration}")
                print(f"outcome: {report.outcome}")
                print(f"err: {report.capstderr}")
                print(f"line_no:{report.location[1]}")
                print(f"test:{report.location[2]}")
                print("## End ##")


def build_dict_tree(items: list[Item]) -> dict:
    def create_node(
        node: Node, parent_name: Node | None = None, parent_type: Node | None = None
    ) -> dict[Any, Any]:
        return {
            "name": node.name,
            "nodeid": node.nodeid,
            "markers": [mark.name for mark in node.own_markers],
            "path": node.path.as_posix(),
            "lineno": node.reportinfo()[1]
            if isinstance(node, Class)
            else (node.location[1] if isinstance(node, Function) else 0),
            "parent_name": parent_name,
            "parent_type": parent_type,
            "type": type(node).__name__.upper(),
            "favourite": False,
            "status": "",
            "children": [],
        }

    def add_node(node_list: list[Node], sub_tree: dict):
        if not node_list:
            return

        # take root node
        current_node = node_list.pop(0)
        node_dict = create_node(
            node=current_node,
            parent_name=current_node.parent.name,
            parent_type=type(current_node.parent).__name__.upper(),
        )

        existing_node = next(
            (
                node
                for node in sub_tree["children"]
                if node["nodeid"] == current_node.nodeid
            ),
            None,
        )

        if existing_node is None:
            sub_tree["children"].append(node_dict)
            existing_node = node_dict

        add_node(
            node_list=node_list,
            sub_tree=existing_node,
        )

    tree: dict[Any, Any] = {}
    root = items[0].listchain()[1]
    tree[root.name] = create_node(node=root)

    for item in items:
        # gets all parents except session
        parts_to_collect = item.listchain()[1:]
        add_node(node_list=parts_to_collect[1:], sub_tree=tree[root.name])

    return {"tree": tree, "meta": {"test_count": len(items)}}
